""" Roach management scene. """

from pygame.constants import BLEND_RGBA_MULT, BLEND_RGBA_SUB
from pgzero.constants import keys, mouse
from pgzero.actor import Actor
from pgzero.screen import Screen
from ..actors.buttons import ImageButton
from ..actors.surf import SurfActor
from ..constants import WIDTH, HEIGHT
from ..roaches import big_roaches, roach_by_name
from ..serums import big_serums, roach_is_serumless, SERUMS
from ..vehicles.base import Vehicle
from .base import Scene, ChangeSceneEvent, defer_to_update
from ..util import safepath


TOOLBAR_LEFT_X = WIDTH * 3 // 4
TOOLBAR_TOP_Y = 0
TOOLBAR_MID_Y = HEIGHT * 1 // 2
VEHICLE_MID_X = WIDTH * 3 // 8
VEHICLE_MID_Y = HEIGHT * 1 // 2
BUTTON_INSET = (20, 20)
SERUM_OFFSET = (0, 20)
ROACH_PAD_OFFSET = (0, 20)


def inset_button(pos, d):
    return (
        pos[0] + d[0] * BUTTON_INSET[0],
        pos[1] + d[1] * BUTTON_INSET[1])


class RoachesScene(Scene):
    """ Roach management scene. """

    def __init__(self, level_scene=None):
        super().__init__()
        self._level_scene = level_scene
        self._vehicle = None
        self._seat_pos = 0
        self._outside_roach = None
        self._outside_roach_pos = 0
        self._inventory_pos = 0
        self._inventory_item = None
        self._roach_layer = self.actors.add_layer("roaches", level=10)
        self._inventory_layer = self.actors.add_layer("inventory", level=10)
        self._pad_layer = self.actors.add_layer("pads", level=5)
        self._seat_layer = self.actors.add_layer("seats", level=5)
        self._button_layer = self.actors.add_layer("buttons", level=6)
        self._init_pads()
        self._init_buttons()
        self._init_serums()

    def enter(self, world):
        self._vehicle = Vehicle.current(world)
        self._init_bg()
        self._init_seats()
        self._init_roaches(world)
        self._update_inventory(world)

    def _init_bg(self):
        self.actors.default.clear()
        overlay = self._vehicle.roach_management_overlay()
        base = overlay.copy()
        if self._level_scene is not None:
            base.fill((0, 0, 0))
            self._level_scene.draw(Screen(base))
        else:
            base.fill((10, 10, 10))
        base.blit(overlay, (0, 0))
        frame = self._vehicle.roach_management_frame()
        if frame is not None:
            frame = frame.copy()
            frame.fill((255, 255, 255, 8), None, BLEND_RGBA_MULT)
            frame_rect = frame.get_rect()
            base.blit(frame, (
                VEHICLE_MID_X - frame_rect.w // 2,
                VEHICLE_MID_Y - frame_rect.h // 2),
                None, BLEND_RGBA_SUB)
        self.actors.default.add(SurfActor(base))

    def _init_seats(self):
        self._seat_layer.clear()
        for seat in self._vehicle.seats:
            seat_actor = self._seat_layer.add(seat.actor())
            seat_actor.pos = (
                seat.vehicle_pos[0] + VEHICLE_MID_X,
                seat.vehicle_pos[1] + VEHICLE_MID_Y)
        self._seat_layer[self._seat_pos].selected = True

    def _init_roaches(self, world):
        self._roach_actors = {}
        for roach in world.roaches:
            self._roach_actors[roach.name] = big_roaches.assemble(roach)

    def _init_serums(self):
        self._serum_actors = {
            serum: big_serums.assemble(serum) for serum in SERUMS}

    def _update_inventory(self, world):
        self._inventory_layer.clear()
        serums = world.serums
        if not serums:
            return
        self._inventory_pos %= len(serums)
        serum_actor = self._serum_actors[serums[self._inventory_pos]]
        inv_pad_centre = self._inventory_pad.center
        serum_actor.pos = (
            inv_pad_centre[0] + SERUM_OFFSET[0],
            inv_pad_centre[1] + SERUM_OFFSET[1])
        self._inventory_layer.add(serum_actor)

    def _update_roaches(self, world):
        self._roach_layer.clear()
        seating = self._vehicle.seating(world)
        self._outside_roaches = []
        for roach in world.roaches:
            seat_pos = seating.get(roach.name)
            if seat_pos is not None:
                roach_actor = self._roach_actors[roach.name]
                roach_actor.pos = self._seat_layer[seat_pos].pos
                self._roach_layer.add(roach_actor)
            else:
                self._outside_roaches.append(roach.name)
        if self._outside_roaches:
            self._outside_roach_pos %= len(self._outside_roaches)
            roach_actor = self._roach_actors[
                self._outside_roaches[self._outside_roach_pos]]
            roach_pad_center = self._roach_pad.center
            roach_actor.pos = (
                roach_pad_center[0] + ROACH_PAD_OFFSET[0],
                roach_pad_center[1] + ROACH_PAD_OFFSET[1])
            self._roach_layer.add(roach_actor)
        else:
            self._outside_roach_pos = 0

    def _init_pads(self):
        self._roach_pad = self._pad_layer.add(
            Actor(safepath("roach_management/roach_pad"), anchor=("left", "bottom")))
        self._roach_pad.pos = (TOOLBAR_LEFT_X, TOOLBAR_MID_Y)
        self._inventory_pad = self._pad_layer.add(
            Actor(safepath("roach_management/inventory_pad"), anchor=("left", "top")))
        self._inventory_pad.pos = (TOOLBAR_LEFT_X, TOOLBAR_MID_Y)

    def _add_button(self, name, anchor, inset, pos, action):
        button = self._button_layer.add(
            ImageButton(name, anchor=anchor, action=action))
        button.pos = inset_button(pos, inset)
        return button

    def _init_buttons(self):
        self._button_layer.clear()
        self._add_button(
            safepath("roach_management/left_button"), ("left", "bottom"), (1, -1),
            self._roach_pad.bottomleft, self._roach_left)

        self._add_button(
            safepath("roach_management/right_button"), ("right", "bottom"), (-1, -1),
            self._roach_pad.bottomright, self._roach_right)

        self._add_button(
            safepath("roach_management/left_button"), ("left", "bottom"), (1, -1),
            self._inventory_pad.bottomleft, self._inventory_left)

        self._add_button(
            safepath("roach_management/right_button"), ("right", "bottom"), (-1, -1),
            self._inventory_pad.bottomright, self._inventory_right)

        self._add_button(
            safepath("roach_management/eject_button"), ("right", "top"), (-1, 1),
            (TOOLBAR_LEFT_X, TOOLBAR_TOP_Y), self._eject_roach)

    def _roach_left(self):
        self._outside_roach_pos -= 1

    def _roach_right(self):
        self._outside_roach_pos += 1

    def _inventory_left(self):
        self._inventory_pos -= 1

    def _inventory_right(self):
        self._inventory_pos += 1

    def _select_seat(self, seat_pos):
        self._seat_layer[self._seat_pos].selected = False
        self._seat_pos = seat_pos
        self._seat_layer[self._seat_pos].selected = True

    @defer_to_update
    def _eject_roach(self, world):
        self._vehicle.seat_roach(world, None, self._seat_pos)

    @defer_to_update
    def _click_roach_pad(self, world):
        if self._outside_roaches:
            roach = self._outside_roaches[self._outside_roach_pos]
            self._vehicle.seat_roach(world, roach, self._seat_pos)

    @defer_to_update
    def _click_inventory_pad(self, world):
        roach_name = self._vehicle.roach_at(world, self._seat_pos)
        if roach_name is None:
            return
        roach = roach_by_name(world, roach_name)
        if roach is None:
            return
        serums = list(world.serums)
        if self._inventory_pos >= len(serums):
            return
        serum = serums.pop(self._inventory_pos)
        if roach_is_serumless(roach):
            roach[serum] = True
            world.serums = serums
            self._update_roach_actor(roach_name)

    @defer_to_update
    def _update_roach_actor(self, world, roach_name):
        roach = roach_by_name(world, roach_name)
        self._roach_actors[roach_name] = big_roaches.assemble(roach)

    def update(self, world, engine, dt):
        super().update(world, engine, dt)
        events = world.pop_events()
        self._update_inventory(world)
        self._update_roaches(world)
        return events

    def on_key_down(self, key, mod, unicode):
        if key in (keys.ESCAPE, keys.Z):
            if self._level_scene is None:
                from .menu import MenuScene
                return [ChangeSceneEvent(MenuScene())]
            return [ChangeSceneEvent(self._level_scene)]

    def on_mouse_down(self, pos, button):
        if button == mouse.LEFT:
            for actor in self.actors.buttons:
                if actor.collidepoint(pos):
                    return actor.action()
            for i, actor in enumerate(self.actors.seats):
                if actor.collidepoint(pos):
                    return self._select_seat(i)
            if self._roach_pad.collidepoint(pos):
                return self._click_roach_pad()
            if self._inventory_pad.collidepoint(pos):
                return self._click_inventory_pad()
